/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Matt Eagleson (e.Messiah@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <gnome.h>
#include <gtk/gtk.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include "uulib/uudeview.h"
#include "uulib/fptools.h"
#include "pan.h"

static pthread_mutex_t online_m = PTHREAD_MUTEX_INITIALIZER;
static gint online = FALSE;
static gint must_exit = FALSE;

void
go_online(void)
{
	pthread_mutex_lock(&online_m); 
    if (online) {
		pthread_mutex_unlock(&online_m);
        return;
    } else {   
        if (connect_to_server()) {
			pthread_mutex_unlock(&online_m);
            online = FALSE;
			must_exit = FALSE;
			queue_purge();
            pthread_exit(NULL);
        } else {
            online = TRUE;
			pthread_mutex_unlock(&online_m);
        }
    }
}

void
go_offline(void)
{
    pthread_mutex_lock(&online_m);
    if (online) {
		put_server(_("Couldn't write to server"), "QUIT\r\n");
	    tcp_close();
	    online = FALSE;
		must_exit = FALSE;
	    server_menu_sensitize();
	    gui_clean();
		queue_purge();
	}
    pthread_mutex_unlock(&online_m);
    pthread_exit(NULL);
}

void
network_error(const gchar * format,...)
{
    va_list args;
    gchar *p = NULL;

    g_return_if_fail(format != NULL);

    va_start(args, format);
    p = g_strdup_vprintf(format, args);
    va_end(args);

    pan_lock();
    gnome_error_dialog(p);
    pan_unlock();
    g_free(p);
    go_offline();
}

gint
set_group(group_data * data)
{
    gchar *buffer;

    g_assert(data);

    put_server_va(_("Error writing to socket."), "GROUP %s\r\n", data->name); 
    buffer = get_server(_("Error reading from socket."));
    if (atoi(buffer) != 211)
        return 1;
    else
        return 0;
}

/***************************************************************
* article_post
****************************************************************/

void
article_post(queue_data * qdata)
{
    gchar *buffer;
    gchar *message;

    go_online();
    message = qdata->str;

    gui_set_status(_("Posting message..."));
    put_server(_("Error writing to socket."), "POST\r\n");
    buffer = get_server(_("Error reading from socket."));
    if (atoi(buffer) == 340) {
        put_server_va(_("Error writing to socket."), "%s\r\n.\r\n", message);
		buffer = get_server(_("Error reading from socket."));
        if (atoi(buffer) == 240) {
          	gui_set_info(_("Message posted.")); 
			gui_clean();
			return;
        }
    }
	gui_clean();
    pan_error_dialog(_("Error posting message.  Server said:\n'%s'"), buffer);
}

/***************************************************************
* grouplist_download 
****************************************************************/

void
grouplist_download(void)
{
    gchar *buffer = NULL;
    group_data *gdata = NULL;
    gint count = 0;

    go_online();

    pan_lock();
    grouplist_free(Pan.grouplist);
    pan_unlock();
    Pan.grouplist = NULL;
    gui_set_status(_("Retrieved 0 groups."));

    put_server(_("Error writing to socket."), "LIST active\r\n");
    buffer = get_server(_("Error reading from socket."));
    if (atoi(buffer) != 215)
        network_error(_("Group list could not be retrieved\nServer said: %s"), buffer);
    else {
        buffer = get_server(_("Error reading from socket."));
        gui_set_progress_activity(1);
        while (strncmp(buffer, ".\r\n", 3)) {
			if (must_exit) goto grouplist_download_cleanup;
            gdata = (group_data *) g_malloc(sizeof(group_data));
    		memset(gdata, 0, sizeof(group_data));
            gdata->name = g_strdup(strtok(buffer, " "));
            Pan.grouplist = g_slist_prepend(Pan.grouplist, gdata);
        	gui_set_progress_activity(++count);
            if ((count % 23) == 0)
                gui_set_status_va(_("Retrieved %d groups."), count);
            buffer = get_server(_("Error reading from socket."));
        }
    }
    gui_set_info_va(_("Retrieved %d groups."), count);
	gui_set_status_va(_("Sorting %d groups."), count);
    Pan.grouplist = g_slist_sort(Pan.grouplist, (GCompareFunc) compare_group_data);
    grouplist_save();
    group_clist_update();
	
	grouplist_download_cleanup:
    gui_clean();
    if (must_exit) {
		grouplist_free(Pan.grouplist);
	}
}

/***************************************************************
* articlelist_download 
****************************************************************/

void
articlelist_download(queue_data * qdata)
{
	gchar *buffer = NULL;
	article_data *adata = NULL;
	gint count = 0, total, mp;
	gint first = 0;
	gint last = 0;
	gint p_last = 0;
	gchar *mid, *part, *parts;
	group_data *gdata = NULL;
	time_t current_time, date;
	GSList *article_list = NULL;
	gchar **sd;

	go_online();
    gdata = qdata->gdata;

    if (qdata->flags & QUEUE_ARTICLELIST_DOWNLOAD_NEW) {
        if (group_index_exists(gdata)) {
            article_list = db_load_articles(gdata);
        }
    }

    gui_set_status(_("Retrieving Article Headers"));

    put_server_va(_("Error writing to socket."), "GROUP %s\r\n", gdata->name);
    buffer = get_server(_("Error reading from socket."));

    if (strtol(buffer, NULL, 10) != 211)
        network_error(_("Couldn't select group '%s'"), gdata->name);

    sscanf(buffer, "%*d %d %d %d", &total, &first, &last);

    p_last = db_fetch_group_last(gdata);
    if (total == '0') {
		gui_set_info(_("No articles"));
        goto articlelist_download_cleanup;
    }
    if ((qdata->flags & QUEUE_ARTICLELIST_DOWNLOAD_NEW) && p_last) {
        if (p_last == last) {
            gui_set_info(_("No new articles"));
            goto articlelist_download_cleanup;
        } 
		put_server_va(_("Error writing to socket."), "XOVER %d-%d\r\n", p_last, last);
    } else {
        if ((qdata->flags & 0x003F) == QUEUE_ARTICLELIST_DOWNLOAD_SAMPLE) {
            put_server_va(_("Error writing to socket"), "XOVER %d-%d\r\n", first, first+50);
        } else {
			put_server_va(_("Error writing to socket."), "XOVER %d-%d\r\n", first, last);
		}
    }

    buffer = get_server(_("Error reading from socket."));
    if (strtol(buffer, (char **) NULL, 10) != 224)
        network_error(_("Couldn't retrieve headers\nServer said: %s"), buffer);
    time(&current_time);
    buffer = get_server(_("Error reading from socket."));
    while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) goto articlelist_download_cleanup;
		sd = g_strsplit(buffer, "\t", 0);
		date = parse_date(sd[3]);
		if ((current_time - date) > (86400 * Pan.props->expiration)) {
        	buffer = get_server(_("Error reading from socket."));
			continue;
		}								 
		adata = (article_data *) g_malloc(sizeof(article_data));
        memset(adata, 0, sizeof(article_data));
        adata->number = g_strdup(sd[0]);
        adata->subject = g_strdup(sd[1]);
        adata->author = g_strdup(sd[2]);
        adata->date = date;
        adata->message_id = g_strdup(sd[4]);
        adata->references = g_strdup(sd[5]);
        adata->linecount = strtol(sd[7], NULL, 10);
        adata->parts = 1;
        adata->part = 1;
        adata->group = gdata;
        
		g_strfreev(sd);
		
		/* Here we look for the (n/N) or [n/N] construct in subject lines */
        mid = adata->subject;
        part = NULL;
        parts = NULL;
        mp = 0;
        while (((int) *(mid) != '\0')) {
            if (((int) *(mid) == '(') || ((int) *(mid) == '[')) {
                if (isdigit((int) *(++mid))) {
                    part = mid;
                    for (; isdigit((int) *(mid)); mid++);
                    if ((int) *(mid) == '/') {
                        if (isdigit((int) *(++mid))) {
                            parts = mid;
                            if (strtol(parts, (char **) NULL, 10) >= strtol(part, (char **) NULL, 10)) {
                                for (; isdigit((int) *(mid)); mid++);
                                if (((int) *(mid) == ')') || ((int) *(mid) == ']')) {
                                    mp = 1;
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            mid++;
        }

        // Handle Multiparts
        if (mp) {
            adata->parts = strtol(parts, (char **) NULL, 10);
            adata->part = strtol(part, (char **) NULL, 10);
        }
        // Add the article to the article list
        article_list = g_slist_prepend(article_list, adata);

        gui_set_progress((gfloat)++count/(gfloat)total);
        buffer = get_server(_("Error reading from socket."));
 	}
	/* Check to see if all articles were expired */
	if (article_list == NULL) {
        gui_set_info(_("All articles were expired"));
		goto articlelist_download_cleanup;
	}

    gui_set_status(_("Saving Article List"));
    
    db_store_group_last(gdata, last);
	gdata->articles += count;

	gui_set_info_va(_("Retrieved %d article headers"), count);
	gui_set_status(_("Threading Article List"));
    gui_set_progress(0);

    article_list = thread_list_by_multipart(article_list, gdata);
	if (must_exit) goto articlelist_download_cleanup;
    article_list = thread_list_by_subject(article_list, gdata);
	if (must_exit) goto articlelist_download_cleanup;
    article_list = thread_list_by_reference(article_list, gdata);
	if (must_exit) goto articlelist_download_cleanup;

	gdata->article_list = article_list;
   	article_list = NULL;
    db_save_articles(gdata);
 	group_clist_update_row(gdata);
	
	if (qdata->flags & QUEUE_FOREGROUND) {
        gui_set_status(_("Displaying article list"));
        article_ctree_update((group_data *) gdata);
    }
	articlelist_download_cleanup:
	g_free(qdata);
	gui_clean();
	if (must_exit) {
		g_slist_free(article_list);
		go_offline();
	}
}

/***************************************************************
* article_download 
***************************************************************/

void
article_download(queue_data * qdata)
{
    gchar *buffer;
    group_data *gdata;
	article_data *adata = NULL;
   	gint count;

	g_assert(qdata->str);

    go_online();
    
    if ((adata = db_fetch_article(qdata->str)) == NULL) {
        gui_set_info_va(_("Couldn't find article header for message_id: %s"), qdata->str);
        goto article_download_cleanup;
    }

	adata->state = STATE_ERROR;
	gdata = qdata->gdata;
    gui_set_status_va(_("Downloading article '%s'"), adata->subject);

    if (set_group(gdata))
        network_error(_("Couldn't select group '%s'"), gdata->name);

    put_server_va(_("Error writing to socket."), "HEAD %s\r\n", adata->number);
    buffer = get_server(_("Error reading from socket."));

    if (strtol(buffer, (char **) NULL, 10) != 221)
        goto article_download_cleanup;

	adata->header = g_strdup("");
	buffer = get_server(_("Error reading from socket."));
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) goto article_download_cleanup;
		adata->header = g_realloc(adata->header, strlen(adata->header) + strlen(buffer) + 1);
		adata->header = strcat(adata->header, buffer);
		buffer = get_server(_("Error reading from socket."));
	}
	
	put_server_va(_("Error writing to socket."), "BODY %s\r\n", adata->number);
    buffer = get_server(_("Error reading from socket."));
    if (strtol(buffer, (char **) NULL, 10) != 222)
        goto article_download_cleanup;
	
	adata->body = g_strdup("");
	count = 0;
	buffer = get_server(_("Error reading from socket."));
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) goto article_download_cleanup;
		adata->body = g_realloc(adata->body, strlen(adata->body) + strlen(buffer) + 1);
		adata->body = strcat(adata->body, buffer);
    	gui_set_progress((gfloat)count++/(gfloat)adata->linecount);
		buffer = get_server(_("Error reading from socket."));
	}
    gui_set_progress(1);

    adata->state = STATE_READ;

    if (qdata->flags & QUEUE_FOREGROUND)
        update_text((article_data *) adata);

    article_download_cleanup:
	g_free(qdata);
	gui_clean();
	if (adata) {
		if (must_exit)
			adata->state = STATE_NONE;
		article_update(adata);
	    article_free(adata);
	}
	if (must_exit)
		go_offline();
}

void
UUD_message_callback(void *opaque, char *msg, int level)
{
    if ((level == UUMSG_WARNING) || (level == UUMSG_ERROR)) {
        pan_error_dialog(_("An error occured during decoding.\n%s"), msg);
    }
}

int
UUD_busy_callback(void *opaque, uuprogress * progress)
{
	gfloat t = (gfloat)1 / (gfloat)3;
		
	switch (progress->action) {
    case UUACT_SCANNING:
        gui_set_progress((gfloat)progress->percent / (gfloat)100 * t);
        return(0);
    case UUACT_DECODING:
        gui_set_progress(((gfloat)progress->percent / (gfloat)100 * t) + t);
        return(0);
    case UUACT_COPYING:
        gui_set_progress(((gfloat)progress->percent / (gfloat)100 * t) + t + t);
        return(0);
    default:
        g_warning("Reached default in UUD_busy_callback\n");
        return(0);
    }
}

char *
UUD_file_name_filter(void *opaque, char *fname)
{
    char *ptr;
    char *name;
    char *p;

    if (fname == NULL)
        return NULL;

    /*
     * strip directory information
     */

    p = tilde_expand(Pan.props->download_dir);

    if ((ptr = _FP_strrchr(fname, '/')) != NULL) {
        name = g_strconcat(p, "/", ptr + 1, NULL);
    } else if ((ptr = _FP_strrchr(fname, '\\')) != NULL) {
        name = g_strconcat(p, "/", ptr + 1, NULL);
    } else {
    	name = g_strconcat(p, "/", fname, NULL);
	}
    g_free(p);
    return name;
}

void
decode(article_data *adata, gchar *filename)
{
	uulist *item;

    UUInitialize();
    UUSetMsgCallback(adata, UUD_message_callback);
    UUSetBusyCallback(adata, UUD_busy_callback, 1);
    UUSetFNameFilter(adata, UUD_file_name_filter);
    if ((UULoadFile(filename, NULL, 0)) != UURET_OK) {
        pan_error_dialog(_("Decoded couldn't load temporary file:\n%s"), filename);
    } else if ((item = UUGetFileListItem(0)) != NULL) {
        if ((item->state & UUFILE_OK) != 0) {
            if ((UUDecodeFile(item, NULL)) != UURET_OK) {
                pan_error_dialog(_("Error decoding '%s'"), adata->subject);
                gui_set_info_va(_("Error decoding '%s'"), adata->subject);
            } else {
                adata->state = STATE_DECODED;
                gui_set_info_va(_("Successfully decoded '%s'"), UUFNameFilter(item->filename));
            }
        } else {
			pan_error_dialog(_("Error decoding '%s'"), adata->subject);
			gui_set_info_va(_("Error decoding '%s'"), adata->subject);
		}
    } else {
		adata->state = STATE_NONE;
	}
	UUCleanUp();
}

/***************************************************************
* article_decode 
***************************************************************/

void
article_decode(queue_data * qdata)
{
    gchar *buffer = NULL;
    group_data *gdata = NULL;
    gint count;
	article_data *adata;
	gchar *p;
	gchar *tmpfile_name;
	FILE *tmpfile;
	
	go_online();
    gdata = qdata->gdata;

    g_assert(qdata->str);
	if ((adata = db_fetch_article(qdata->str)) == NULL) {
		gui_set_info_va(_("Couldn't find article header for message_id: %s"), qdata->str);
		return;
    }
    
    adata->state = STATE_ERROR;

    p = tilde_expand(Pan.props->temp_dir);
    tmpfile_name = tempnam(p, "pan");
	g_free(p);
    tmpfile = fopen(tmpfile_name, "w");
	
    if (tmpfile == NULL)
        network_error(_("Couldn't open temporary file %s"), tmpfile_name);
    
    if (set_group(gdata))
        network_error(_("Couldn't select group '%s'"), gdata->name);
       	
	gui_set_status_va(_("Downloading article body '%s'"), adata->subject);
    gui_set_progress(0);
    
	put_server_va(_("Error writing to socket."), "BODY %s\r\n", adata->number);
    buffer = get_server(_("Error reading from socket."));
    if (strtol(buffer, NULL, 10) != 222)
		goto article_decode_cleanup;

	count = 1;
	buffer = get_server(_("Error reading from socket."));
	while (strncmp(buffer, ".\r\n", 3)) {
		if (must_exit) goto article_decode_cleanup;
		fputs(buffer, tmpfile);
		buffer = get_server(_("Error reading from socket."));
		gui_set_progress(percent(count++, adata->linecount));
	}
	gui_set_progress(1);
	fclose(tmpfile);
	tmpfile = NULL;
	
	gui_set_status_va(_("Decoding article: %s"), adata->subject);
    gui_set_progress(0);

    decode(adata, tmpfile_name);

	article_decode_cleanup:
	g_free(qdata);
	gui_clean();
	if (adata)	{
		if (must_exit)
			adata->state = STATE_NONE;
		article_update(adata);
    	article_free(adata);
	}
    if (tmpfile_name) {
		unlink(tmpfile_name);
		g_free(tmpfile_name);
	}
    if (tmpfile)
		fclose(tmpfile);
	if (must_exit)
		go_offline();
}

/***************************************************************
* multipart_decode 
***************************************************************/

void
multipart_decode(queue_data * qdata)
{
    gchar *buffer;
    group_data *gdata;
    gint i = 0;
    gint count;
	article_data *fadata = NULL;
	article_data *adata = NULL;
	gchar *p;
	gchar *tmpfile_name;
	FILE *tmpfile;
	gchar **sd;

    go_online();
    
    gdata = qdata->gdata;

    sd = g_strsplit(qdata->str, "\n", -1);
    g_assert(sd);

    /* Find the (fake) article header for the
     * multipart node */
    if ((fadata = db_fetch_article(sd[2])) == NULL) {
		gui_set_info_va(_("Couldn't find article header for message_id: %s"), qdata->str);
        return;
    }

	/* Set state to error until we are proved otherwise
	 * (Somewhat pessimistic isn't it?) */
	fadata->state = STATE_ERROR;

    p = tilde_expand(Pan.props->temp_dir);
    tmpfile_name = tempnam(p, "pan");
	g_free(p);

    if ((tmpfile = fopen(tmpfile_name, "w")) == NULL) {
        gui_set_info(_("Multipart decode failed: Couldn't open tempfile"));
        goto multipart_decode_cleanup;
    }
    if (set_group(gdata))
        network_error(_("Couldn't select group '%s'"), gdata->name);
    for (i = 2; sd[i]; i++) {
		adata = NULL;
		if ((adata = db_fetch_article(sd[i])) == NULL) {
			gui_set_info_va(_("Couldn't find article header for message_id: %s"), sd[i]);
            goto multipart_decode_cleanup;
        }
		gui_set_status_va(_("Downloading part %d of %d: %s"), adata->part, adata->parts, adata->subject);
        gui_set_progress(0);

		put_server_va(_("Error writing to socket."), "body %s\r\n", adata->number);
        buffer = get_server(_("Error reading from socket."));

        if (strtol(buffer, (char **) NULL, 10) != 222)
            goto multipart_decode_cleanup;

        count = 1;
        buffer = get_server(_("Error reading from socket."));

        while (strncmp(buffer, ".\r\n", 3)) {
			if (must_exit) goto multipart_decode_cleanup;
            fputs(buffer, tmpfile);
            buffer = get_server(_("Error reading from socket."));
            gui_set_progress(percent(count++,adata->linecount));
        }
        article_free(adata);
    }
	adata = NULL;
	fclose(tmpfile);
	tmpfile = NULL;

    gui_set_progress(1);
    gui_set_status_va(_("Decoding multipart '%s'"), sd[1]);
    gui_set_progress(0);
	g_strfreev(sd);

    decode(fadata, tmpfile_name);

	multipart_decode_cleanup:
	g_free(qdata);
	gui_clean();
	if (fadata) {
		if (must_exit)
			fadata->state = STATE_NONE;
		article_update(fadata);
    	article_free(fadata);
	}
    if (tmpfile_name) {
		unlink(tmpfile_name);
		g_free(tmpfile_name);
	}
    if (tmpfile)
		fclose(tmpfile);
	if (adata)
		article_free(adata);
	if (must_exit) {
		go_offline();
	}
}

/*
void
send_keepalive()
{
    gchar *buffer;

    if (online_check())
        return;

    if (put_server("help\r\n") == -1)
        network_error(_("Error writing to socket."));
    if ((buffer = get_server()) == NULL)
        network_error(_("Error reading from socket."));
    if (strtol(buffer, (char **) NULL, 10) == 100) {
        while (strncmp(buffer, ".\r\n", 3)) {
            if ((buffer = get_server()) == NULL)
                network_error(_("Error reading from socket."));
        }
        return;
    } else {
        network_error(_("Problem with HELP no-op."));
    }
    return;
}
*/

int
connect_to_server(void)
{
    gchar *buffer;
    gint response;

    if (Pan.active_server == NULL) {
        pan_error_dialog(_("There is no active server.\nPlease configure one."));
        return(1);
    }
    gui_set_status(_("Connecting to server"));

    if (tcp_open(Pan.active_server, "nntp") < 0) {
        gui_clean();
        pan_error_dialog(_("Could not connect to server."));
        return(1);
    }
    if ((buffer = get_server(NULL)) == NULL) {
        gui_clean();
        pan_error_dialog(_("Could not connect to server."));
        return(1);
    }
    gui_set_status_default(_("Pan - Online"));

    response = strtol(buffer, (char **) NULL, 10);
    if ((response != 201) && (response != 200)) {
        gui_clean();
        pan_error_dialog(_("Could not connect to server."));
        return(1);
    }
    if (response == 201)
        Pan.active_server->posting = 0;
    else
        Pan.active_server->posting = 1;
    return(0);
}

void
stop_all_online_tasks(void)
{
	pthread_mutex_lock(&online_m);
    if (online)
        must_exit = TRUE;
	pthread_mutex_unlock(&online_m);
}
